from pathlib import Path
from tempfile import NamedTemporaryFile
import pytest
from click.testing import CliRunner
from broker import Broker
from broker.commands import cli
from broker.providers.beaker import Beaker
from broker.settings import inventory_path, settings_path

SCENARIO_DIR = Path("tests/data/cli_scenarios/beaker")


@pytest.fixture(scope="module", autouse=True)
def skip_if_not_configured():
    try:
        Beaker()
    except Exception as err:
        pytest.skip(f"Beaker is not configured correctly: {err}")


@pytest.fixture(scope="module")
def temp_inventory():
    """Temporarily move the local inventory, then move it back when done"""
    backup_path = inventory_path.rename(f"{inventory_path.absolute()}.bak")
    yield
    CliRunner().invoke(cli, ["checkin", "--all", "--filter", '@inv._broker_provider == "Beaker"'])
    inventory_path.unlink()
    backup_path.rename(inventory_path)


@pytest.mark.parametrize(
    "args_file", [f for f in SCENARIO_DIR.iterdir() if f.name.startswith("checkout_")], ids=lambda f: f.name.split(".")[0]
)
def test_checkout_scenarios(args_file, temp_inventory):
    result = CliRunner().invoke(cli, ["checkout", "--args-file", args_file])
    assert result.exit_code == 0


# @pytest.mark.parametrize(
#     "args_file", [f for f in SCENARIO_DIR.iterdir() if f.name.startswith("execute_")], ids=lambda f: f.name.split(".")[0]
# )
# def test_execute_scenarios(args_file):
#     result = CliRunner().invoke(cli, ["execute", "--args-file", args_file])
#     assert result.exit_code == 0


def test_inventory_sync():
    result = CliRunner().invoke(cli, ["inventory", "--sync", "Beaker"])
    assert result.exit_code == 0


def test_jobs_list():
    result = CliRunner(mix_stderr=False).invoke(cli, ["providers", "Beaker", "--jobs", "--mine"])
    assert result.exit_code == 0


# def test_job_query():
#     """This isn't possible until we can figure out how to capture logged output"""
#     result = CliRunner().invoke(
#         cli, ["providers", "Beaker", "--job", "<job_id>"]
#     )
#     assert result.exit_code == 0


# ----- Broker API Tests -----


def test_beaker_host():
    with Broker(job_xml="tests/data/beaker/test_job.xml") as r_host:
        res = r_host.execute("hostname")
        assert res.stdout.strip() == r_host.hostname
        remote_dir = "/tmp/fake"
        r_host.session.sftp_write(str(settings_path.absolute()), f"{remote_dir}/")
        res = r_host.execute(f"ls {remote_dir}")
        assert str(settings_path.name) in res.stdout
        with NamedTemporaryFile() as tmp:
            r_host.session.sftp_read(f"{remote_dir}/{settings_path.name}", tmp.file.name)
            data = r_host.session.sftp_read(f"{remote_dir}/{settings_path.name}", return_data=True)
            assert (
                settings_path.read_bytes() == Path(tmp.file.name).read_bytes()
            ), "Local file is different from the received one"
            assert (
                settings_path.read_bytes() == data
            ), "Local file is different from the received one (return_data=True)"
            assert data == Path(tmp.file.name).read_bytes(), "Received files do not match"
        # test the tail_file context manager
        tailed_file = f"{remote_dir}/tail_me.txt"
        r_host.execute(f"echo 'hello world' > {tailed_file}")
        with r_host.session.tail_file(tailed_file) as tf:
            r_host.execute(f"echo 'this is a new line' >> {tailed_file}")
        assert "this is a new line" in tf.stdout
        assert "hello world" not in tf.stdout



"""Module providing classes to establish ssh or ssh-like connections to hosts.

Classes:
    Session - Wrapper around ssh2-python's auth/connection system.
    InteractiveShell - Wrapper around ssh2-python's non-blocking channel system.
    ContainerSession - Wrapper around docker-py's exec system.

Note: You typically want to use a Host object instance to create sessions,
      not these classes directly.
"""
from contextlib import contextmanager
from pathlib import Path
import tempfile

from logzero import logger

from broker import helpers
from broker.exceptions import NotImplementedError
from broker.settings import settings

SSH_BACKENDS = ("ssh2-python", "ssh2-python312", "ansible-pylibssh", "hussh")
SSH_BACKEND = settings.SSH_BACKEND

logger.debug(f"{SSH_BACKEND=}")


SSH_NOT_SUPPORTED_MSG = (
    f"SSH backend {SSH_BACKEND!r} not supported.\nSupported ssh backends:\n{SSH_BACKENDS}"
)
SSH_NOT_INSTALLED_MSG = (
    f"{SSH_BACKEND} is not installed.\n"
    "ssh actions will not work.\n"
    f"To use ssh, run 'pip install broker[{SSH_BACKEND}]'."
)
SSH_IMPORT_MSG = ""

try:
    if SSH_BACKEND == "ansible-pylibssh":
        from broker.binds.pylibssh import InteractiveShell, Session
    elif SSH_BACKEND == "hussh":
        from broker.binds.hussh import Session
    elif SSH_BACKEND in ("ssh2-python", "ssh2-python312"):
        from broker.binds.ssh2 import InteractiveShell, Session  # noqa: F401
    else:
        SSH_IMPORT_MSG = SSH_NOT_SUPPORTED_MSG
except ImportError:
    SSH_IMPORT_MSG = SSH_IMPORT_MSG or SSH_NOT_INSTALLED_MSG
finally:
    if SSH_IMPORT_MSG:
        logger.warning(SSH_IMPORT_MSG)

        class Session:
            """Default wrapper around ssh backend's auth/connection system."""

            def __init__(self, **kwargs):
                """Initialize a Session object."""


class ContainerSession:
    """An approximation of ssh-based functionality from the Session class."""

    def __init__(self, cont_inst, runtime=None):
        self._cont_inst = cont_inst
        if not runtime:
            runtime = settings.CONTAINER.runtime
        self.runtime = runtime

    def run(self, command, demux=True, **kwargs):
        """Container approximation of Session.run."""
        kwargs.pop("timeout", None)  # Timeouts are set at the client level
        kwargs["demux"] = demux
        if "'" in command:
            with tempfile.NamedTemporaryFile(mode="w", suffix=".sh") as tmp:
                tmp.write(command)
                tmp.seek(0)
                command = f"/bin/bash {tmp.name}"
                self.sftp_write(tmp.name)
        if any(s in command for s in "|&><"):
            # Containers don't handle pipes, redirects, etc well in a bare exec_run
            command = f"/bin/bash -c '{command}'"
        result = self._cont_inst._cont_inst.exec_run(command, **kwargs)
        if demux:
            result = helpers.Result.from_duplexed_exec(result, self.runtime)
        else:
            result = helpers.Result.from_nonduplexed_exec(result)
        return result

    def disconnect(self):
        """Needed for simple compatibility with Session."""

    @contextmanager
    def tail_file(self, filename):
        """Simulate tailing a file on the remote host."""
        initial_size = int(self.run(f"stat -c %s {filename}").stdout.strip())
        yield (res := helpers.Result())
        # get the contents of the file from the initial size to the end
        result = self.run(f"tail -c +{initial_size} {filename}")
        res.__dict__.update(result.__dict__)

    def sftp_write(self, source, destination=None, ensure_dir=True):
        """Add one of more files to the container."""
        # ensure source is a list of Path objects
        if not isinstance(source, list):
            source = [Path(source)]
        else:
            source = [Path(src) for src in source]
        # validate each source's existenence
        for src in source:
            if not Path(src).exists():
                raise FileNotFoundError(src)
        destination = destination or f"{source[0].parent}/"
        # Files need to be added to a tarfile
        with helpers.temporary_tar(source) as tar:
            logger.debug(f"{self._cont_inst.hostname} adding file(s) {source} to {destination}")
            if ensure_dir:
                if destination.endswith("/"):
                    self.run(f"mkdir -m 666 -p {destination}")
                else:
                    self.run(f"mkdir -m 666 -p {Path(destination).parent}")
            self._cont_inst._cont_inst.put_archive(str(destination), tar.read_bytes())

    def sftp_read(self, source, destination=None, return_data=False):
        """Get a file or directory from the container."""
        destination = Path(destination or source)
        logger.debug(f"{self._cont_inst.hostname} getting file {source}")
        data, status = self._cont_inst._cont_inst.get_archive(source)
        logger.debug(f"{self._cont_inst.hostname}: {status}")
        data = b"".join(d for d in data)
        if destination.name == "_raw":
            return data
        with helpers.data_to_tempfile(data, as_tar=True) as tar:
            del data
            if len(tar.getmembers()) == 1:
                f = tar.extractfile(tar.getmember(Path(source).name))
                if return_data:
                    logger.debug(f"Extracting {source}")
                    return f.read()
                else:
                    logger.debug(f"Extracting {source} to {destination}")
                    destination.parent.mkdir(parents=True, exist_ok=True)
                    destination.write_bytes(f.read())
            else:
                logger.warning("More than one member was found in the tar file.")
                tar.extractall(destination.parent if destination.is_file() else destination)

    def shell(self, pty=False):
        """Create and return an interactive shell instance."""
        raise NotImplementedError("ContainerSession.shell has not been implemented")



"""Broker settings module.

Useful items:
    settings: The settings object.
    init_settings: Function to initialize the settings file.
    validate_settings: Function to validate the settings file.
    INTERACTIVE_MODE: Whether or not Broker is running in interactive mode.
    BROKER_DIRECTORY: The directory where Broker looks for its files.
    settings_path: The path to the settings file.
    inventory_path: The path to the inventory file.
"""
import inspect
import os
from pathlib import Path

import click
from dynaconf import Dynaconf, Validator
from dynaconf.validator import ValidationError

from broker.exceptions import ConfigurationError


def init_settings(settings_path, source, interactive=False, is_url=False):
    """Initialize the broker settings file."""
    proceed = not False
    if interactive:
        try:
            proceed = (
                click.prompt(
                    f"Get example file from {source}?\n{source}",
                    type=click.Choice(["y", "n"]),
                    default="y",
                )
                == "y"
            )
        except click.core.Abort:
            # We're likely in a different non-interactive environment (container?)
            global INTERACTIVE_MODE
            proceed, INTERACTIVE_MODE = True, False
    if proceed:
        # get example file from source
        if is_url:
            import requests

            click.echo(f"Downloading example file from: {source}")
            raw_file = requests.get(source, timeout=60)
            settings_path.write_text(raw_file.text)
        else:
            example_file = source.read_text()
            settings_path.write_text(example_file)
        if INTERACTIVE_MODE:
            try:
                click.edit(filename=str(settings_path.absolute()))
            except click.exceptions.ClickException:
                click.secho(
                    f"Please edit the file {settings_path.absolute()} and add your settings.",
                    fg="yellow",
                )
        return True


def init_settings_from_github(settings_path, interactive=False):
    """Initialize the broker settings file."""
    raw_url = (
        "https://raw.githubusercontent.com/SatelliteQE/broker/master/broker_settings.yaml.example"
    )
    return init_settings(settings_path, raw_url, interactive, is_url=True)


def init_settings_from_local_repo(settings_path, interactive=False):
    """Initialize the broker settings file."""
    example_path = Path(__file__).parent.parent.joinpath("broker_settings.yaml.example")
    if not example_path.exists():
        return
    return init_settings(settings_path, example_path, interactive)


INTERACTIVE_MODE = False
# GitHub action context
if "GITHUB_WORKFLOW" not in os.environ:
    # determine if we're being ran from a CLI
    for frame in inspect.stack()[::-1]:
        if "/bin/broker" in frame.filename:
            INTERACTIVE_MODE = True
            break


BROKER_DIRECTORY = Path.home().joinpath(".broker")

if "BROKER_DIRECTORY" in os.environ:
    envar_location = Path(os.environ["BROKER_DIRECTORY"])
    if envar_location.is_dir():
        BROKER_DIRECTORY = envar_location

# ensure the broker directory exists
BROKER_DIRECTORY.mkdir(parents=True, exist_ok=True)

settings_path = BROKER_DIRECTORY.joinpath("broker_settings.yaml")
inventory_path = BROKER_DIRECTORY.joinpath("inventory.yaml")

if not settings_path.exists():
    click.secho(f"Broker settings file not found at {settings_path.absolute()}.", fg="red")
    if not (success := init_settings_from_local_repo(settings_path, interactive=INTERACTIVE_MODE)):
        success = init_settings_from_github(settings_path, interactive=INTERACTIVE_MODE)
    if not success:
        raise ConfigurationError(f"Broker settings file not found at {settings_path.absolute()}.")

validators = [
    Validator("HOST_USERNAME", default="root"),
    Validator("HOST_PASSWORD", default="toor"),
    Validator("HOST_CONNECTION_TIMEOUT", default=60),
    Validator("HOST_SSH_PORT", default=22),
    Validator("HOST_SSH_KEY_FILENAME", default=None),
    Validator("HOST_IPV6", default=False),
    Validator("HOST_IPV4_FALLBACK", default=True),
    Validator("SSH_BACKEND", default="ssh2-python312"),
    Validator("LOGGING", is_type_of=dict),
    Validator(
        "LOGGING.CONSOLE_LEVEL",
        is_in=["error", "warning", "info", "debug", "trace", "silent"],
        default="info",
    ),
    Validator(
        "LOGGING.FILE_LEVEL",
        is_in=["error", "warning", "info", "debug", "trace", "silent"],
        default="debug",
    ),
]

# temporary fix for dynaconf #751
vault_vars = {k: v for k, v in os.environ.items() if "VAULT_" in k}
for k in vault_vars:
    del os.environ[k]

settings = Dynaconf(
    settings_file=str(settings_path.absolute()),
    ENVVAR_PREFIX_FOR_DYNACONF="BROKER",
    validators=validators,
)
# to make doubly sure, remove the vault loader if set somehow
settings._loaders = [loader for loader in settings._loaders if "vault" not in loader]

try:
    settings.validators.validate()
except ValidationError as err:
    raise ConfigurationError(
        f"Configuration error in {settings_path.absolute()}: {err.args[0]}"
    ) from err

os.environ.update(vault_vars)
